!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2015, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
module mod_grid
use mod_global, only : dp
use mod_box, only : dimn
implicit none
save 
private
public read_grid,      &
       set_grid,       &
       ngrid,          &
       deallocate_grid,&
       gpt
real(kind=dp),dimension(:),allocatable :: gpos
real(kind=dp),dimension(:,:),allocatable :: gpt
real(kind=dp) :: glenx,gleny,glenz
character(len=80) :: gfilename
integer :: gdivx,gdivy,gdivz,ngrid
logical :: gfit,gfile,gdef

contains

    subroutine deallocate_grid
    implicit none
    deallocate(gpt)
    deallocate(gpos)
    end subroutine deallocate_grid

    subroutine read_grid
    use mod_output, only : string,output,erroroutput,nerror,error,heading
    use mod_global, only : cmdunit
    use mod_qclib, only : next,downcase,freein
    use mod_box, only : dimn
    implicit none
    !-------------------------------------------------------------------------------
    ! local variables
    !
    character(len=80) :: input
    character(len=5) :: cc
    integer :: lower,upper,i,ldif,idum
    logical :: terminate
    real(kind=dp) :: dum,readin
    !-------------------------------------------------------------------------------
    ! allow indentation with spaces in input file
    !
    heading=.true.
    write(string,'(a)')'**Reading grid data**'; call output
    terminate=.false.
    do
        read(cmdunit,'(a80)') input
        i=1
        do while ((input(i:i).eq.' ').and.(i.le.80))
            i=i+1
        enddo
        upper = i-1
        
        lower = upper
        upper = next(lower,input)
        ldif = upper - lower - 1
        if (ldif.eq.0) then
            cc = '    '
        else
            cc = input(lower+1:upper-1)
        end if
        
        lower=upper
        upper=next(lower,input)
        !-------------------------------------------------------------------------------
        ! do i=lower,upper
        !      if ((input(i:i).eq.'T').or.(input(i:i).eq.'t')) then
        !         flag=.true.
        !         exit
        !      else if ((input(i:i).eq.'F').or.(input(i:i).eq.'f')) then
        !         flag=.false.
        !         exit
        !      endif
        ! enddo
        !
        
        
        select case(downcase(cc))

        case('gfit')
            allocate(gpos(dimn))
            gpos = 0._dp
            gfit = .true.
            ! Set gdivx,gdivy,gdivz
            call freein(input,lower,upper,gdivx,dum,1)
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,gdivy,dum,1)
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,gdivz,dum,1)

            if (dimn .eq. 0) then
                gdivx = 0
                gdivy = 0
                gdivz = 0
            endif
            gfilename='grid.data'
            write(string,3000)  gfit; call output

        case('gdef')
            gdef = .true.
            ! Read gpos
            call freein(input,lower,upper,idum,readin,2)
            allocate(gpos(dimn))
            gpos = 0._dp
            do i=1,dimn
                gpos(i) = readin
                lower=upper
                upper=next(lower,input)
                call freein(input,lower,upper,idum,readin,2)
            enddo

            ! Set glenx,gleny,glenz
            glenx = readin
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,idum,gleny,2)
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,idum,glenz,2)
            lower=upper
            upper=next(lower,input)

            ! Set gdivx,gdivy,gdivz
            call freein(input,lower,upper,gdivx,dum,1)
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,gdivy,dum,1)
            lower=upper
            upper=next(lower,input)
            call freein(input,lower,upper,gdivz,dum,1)
            if (dimn .eq. 0) then
                gdivx = 0
                gdivy = 0
                gdivz = 0
            endif
            gfilename='grid.data'
            write(string,3010)  gpos; call output
            write(string,3020) glenx; call output
            write(string,3030) gleny; call output
            write(string,3040) glenz; call output
            write(string,3050) gdivx; call output
            write(string,3060) gdivy; call output
            write(string,3070) gdivz; call output

        case('gfile')
            gfile = .true.
            gfilename= trim(input(lower+1:upper-1)) // '.data'
            write(string,3001) gfile; call output
        case('end')
            terminate=.true.
        case default
            if (cc .ne.'    ' .and. cc .ne.'%') then
                nerror = 1
                error(1) = 'macro "' // cc // '" not recognized**'
                call erroroutput
                stop
            endif
        end select
        
        if (terminate) exit
    enddo

    3001 format('Read grid from file    = ',l5)
    3000 format('Fit the grid           = ',l5)
    3010 format('Position of the grid   = ',3f10.5)
    3020 format('Length of grid : x-dir = ',f10.5)
    3030 format('                 y-dir = ',f10.5)
    3040 format('                 z-dir = ',f10.5)
    3050 format('Number of divs : x-dir = ',I10)
    3060 format('                 y-dir = ',I10)
    3070 format('                 z-dir = ',I10)
    end subroutine read_grid

    subroutine set_grid(avgsize)
    use mod_box, only : dimn,nat,samplesize
    use mod_particles, only : at
    use mod_output, only : string, output,heading,erroroutput,nerror,error
    use mod_file, only : FileExists,iofile,GetFreeUnit
    implicit none

    ! Passed variables
    real(kind=dp), intent(in) :: avgsize
    ! Local variables
    integer :: i,j,k,l,gunit,maxlines,ios
    real(kind=dp) :: delx,dely,delz
    real(kind=dp) gpos_min(dimn),gpos_max(dimn)
    character(len=1) :: junk

    heading=.true.
    write(string,'(a)')'**Setting the grid**'; call output
    

    if (gfit .eqv. .true. .or. gdef .eqv. .true.) then
        gunit = GetFreeUnit()
        open(gunit,file=gfilename,status='replace',form='formatted')
        write(string,'(a,a)') 'Writing grid points to file ',&
                            trim(gfilename); call output
        delx = 0._dp
        dely = 0._dp
        delz = 0._dp

        if (gfit .eqv. .true.) then
            gpos_min = samplesize
            gpos_max = -samplesize
            do i = 1,nat
                where(at(i)%x .lt. gpos_min)
                    gpos_min = at(i)%x
                end where
                where(at(i)%x .gt. gpos_max)
                    gpos_max = at(i)%x
                end where
            enddo
            gpos_min = gpos_min + avgsize*(/(1._dp,i=1,dimn)/)
            gpos_max = gpos_max - avgsize*(/(1._dp,i=1,dimn)/)

            gpos = gpos_min

            if (gdivx .eq. 0) gpos(1) = 0.5_dp*(gpos_min(1)+gpos_max(1))
            if (gdivy .eq. 0) gpos(2) = 0.5_dp*(gpos_min(2)+gpos_max(2))
            if (gdivz .eq. 0) gpos(3) = 0.5_dp*(gpos_min(3)+gpos_max(3))


            if (gdivx .ne. 0) delx = (gpos_max(1)-gpos_min(1))/gdivx
            if (gdivy .ne. 0) dely = (gpos_max(2)-gpos_min(2))/gdivy
            if (gdivz .ne. 0) delz = (gpos_max(3)-gpos_min(3))/gdivz

        endif

        if (gdef .eqv. .true.) then
            if (glenx .eq. 0._dp) gdivx = 0
            if (gleny .eq. 0._dp) gdivy = 0
            if (glenz .eq. 0._dp) gdivz = 0

            if (gdivx .ne. 0) delx = glenx*samplesize(1)/gdivx
            if (gdivy .ne. 0) dely = gleny*samplesize(2)/gdivy
            if (gdivz .ne. 0) delz = glenz*samplesize(3)/gdivz
        endif

        ngrid = (gdivx+1)*(gdivy+1)*(gdivz+1)
        write(string,3080) ngrid; call output

        allocate(gpt(dimn,ngrid))
        l = 1
        do i = 1,gdivx+1
            do j = 1,gdivy+1
                do k = 1,gdivz+1
                    gpt(:,l) = gpos + (/(i-1)*delx,  &
                                        (j-1)*dely,  &
                                        (k-1)*delz/)
                    write(gunit,*) gpt(:,l)
                    l = l+1
                enddo
            enddo
        enddo

    endif

    if (gfile .eqv. .true.) then
        if (FileExists(gfilename,.false.)) then
            call iofile(gfilename,'formatted  ',gunit,.false.,.false.)
            heading=.true.
            write(string,'(a)') '**Reading grid data from file**'
            call output
        else   
            nerror = 2
            error(1) = 'grid data file not found'
            error(2) = 'in subroutine set_Grid,  Filename = ' &
                       // trim(gfilename)
            call erroroutput
            stop
         endif

        write(string,'(a,a)') 'Reading grid points from file ',&
                            trim(gfilename); call output
        l=0
        maxlines = 100000
        do i = 1,maxlines
            read(gunit,*,iostat=ios) junk
            if (ios .ne. 0) then
                exit
            endif

            if (i .eq. maxlines) then
                nerror = 1
                error(1) = 'Maximum lines reached'
                call erroroutput
                stop
            endif
            l = l+1
        enddo

        rewind(gunit)
        allocate(gpt(dimn,l))
        ngrid = l
        do i = 1,l
            read(gunit,*) gpt(:,i)
        enddo

    endif

    3080 format('Number of grid points  = ',I10)
    end subroutine set_grid
end module mod_grid



