!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2015, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!****************************************************************************
!**
!**   MODULE mod_io: Handle input and output
!**
!**   type definitions:
!**   ----------------
!**   N/A
!**
!**   Real Variable Definitions:
!**   -------------------------------------
!**   samplesize(:)
!**
!**   Integer Variable Definitions:
!**   ----------------------------
!**   N/A
!**  
!**   logical Variable Definitions:
!**   ----------------------------
!**   N/A
!**
!**   Contains Routines:
!**   ------------------
!**   read_input : reads in parameters given by the user
!****************************************************************************
module mod_io
implicit none
save
private
public read_input
contains
    !-------------------------------------------------------------------------------
    ! read_input  :  reads the input file
    !
    subroutine read_input
    use mod_box, only : allocate_box, allocate_species
    use mod_particles, only : read_config
    use mod_output, only : erroroutput,nerror,error
    use mod_global, only : cmdunit
    use mod_qclib, only : next,downcase
    implicit none
    !-------------------------------------------------------------------------------
    ! local variables
    !
    character(len=80) :: filename,input
    character(len=4) :: cc
    integer :: lower,upper,i,ldif
    logical :: terminate

    !-------------------------------------------------------------------------------
    ! allow indentation with spaces in input file
    !
    terminate=.false.
    do
        read(cmdunit,'(a80)') input
        i=1
        do while ((input(i:i).eq.' ').and.(i.le.80))
            i=i+1
        enddo
        upper = i-1
        
        lower = upper
        upper = next(lower,input)
        ldif = upper - lower - 1
        if (ldif.eq.0) then
            cc = '    '
        else
            cc = input(lower+1:upper-1)
        end if

        lower=upper
        upper=next(lower,input)
        filename = trim(input(lower+1:upper-1)) // '.data'

        select case(downcase(cc))
        case('conf')
            call allocate_box(filename)
            ! set a flag that dimensions and pbc are read in
            call read_config(filename)
        case('spec')
            call allocate_species(filename)
        case('end')
            terminate=.true.
        case default
            if (cc .ne.'    ' .and. cc .ne.'%') then
                nerror = 1
                error(1) = 'macro "' // cc // '" not recognized**'
                call erroroutput
                stop
            endif
        end select

        if (terminate) exit
    enddo

    end subroutine read_input
end module mod_io
