!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2015, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!-------------------------------------------------------------------------------
! MODULE mod_rigidity :
!
! type definitions:
! ----------------
! N/A
!
! Real Variable Definitions:
! -------------------------------------
!
! Integer Variable Definitions:
! ----------------------------
!
! logical Variable Definitions: N/A
! ----------------------------
!
! Contains Routines:
! ------------------
!

module mod_rigidity
use mod_global, only : dp
implicit none
save
private
public set_rigidity,project_forces,proj_fij,project,deallocate_rigidity

integer nrow, ncol, nnzmax
real(kind=dp),dimension(:,:),allocatable :: proj_fij
real(kind=dp),dimension(:),allocatable :: proj_fij_scalar,fij_scalar
real(kind=dp), dimension(:), allocatable :: rigidity
integer, dimension(:), allocatable :: rowptr,colind
logical :: project

contains

    subroutine set_rigidity
    !-------------------------------------------------------------------------------
    ! SUBROUTINE set_interatomic :
    !
    ! Passed Parameters :-
    ! -----------------
    !
    ! Non-Obvious Parameter :-
    ! ---------------------
    !
    use mod_global, only : cmdunit
    use mod_output, only : string,output,erroroutput,nerror,error,heading
    use mod_qclib, only : next,downcase
    implicit none
    !-------------------------------------------------------------------------------
    ! local variables
    !
    character(len=80) :: input
    character(len=4) :: cc
    integer :: lower,upper,i,ldif
    logical :: setflag,flag,terminate
    !-------------------------------------------------------------------------------
    ! allow indentation with spaces in input file
    !
    heading=.true.
    write(string,'(a)')'**Reading rigidity data**'; call output
    terminate=.false.
    do
        setflag=.false.
        read(cmdunit,'(a80)') input
        i=1
        do while ((input(i:i).eq.' ').and.(i.le.80))
            i=i+1
        enddo
        upper = i-1
        
        lower = upper
        upper = next(lower,input)
        ldif = upper - lower - 1
        if (ldif.eq.0) then
            cc = '    '
        else
            cc = input(lower+1:upper-1)
        end if
        
        lower=upper
        upper=next(lower,input)
        do i=lower,upper
             if ((input(i:i).eq.'T').or.(input(i:i).eq.'t')) then
                flag=.true.
                setflag = .true.
                exit
             else if ((input(i:i).eq.'F').or.(input(i:i).eq.'f')) then
                flag=.false.
                setflag = .true.
                exit
             endif
        enddo

        select case(downcase(cc)) 
        case('proj')
            if (setflag .eqv. .false.) then
                nerror = 1
                error(1) = 'Unrecognized flag in the uniqueness stage**'
                call erroroutput
                stop
            endif
            project = flag
        case('end')
            terminate=.true.
        case default
            if (cc .ne.'    ' .and. cc .ne.'%') then
                nerror = 1
                error(1) = 'macro "' // cc // '" not recognized**'
                call erroroutput
                stop
            endif
        end select

        if (terminate) exit
    enddo

    write(string,4020) project; call output

    4020 format('Projection = ',L10)

    end subroutine set_rigidity

    subroutine construct_rigidity
    use mod_output, only      : output,erroroutput
    use mod_box, only         : dimn,nat
    use mod_neigh, only       : list_length,marker1,marker2,list
    use mod_particles, only   : at
    use mod_box,only          : samplesize,pbc
    implicit none

    ! Local variables
    integer row,i,j,k,nz
    logical bdry(dimn)
    real(kind=dp) dist_ij
    real(kind=dp) :: rij(dimn)

    nrow=list_length
    ncol=dimn*(nat+1)
    nnzmax=3*dimn*list_length
    
    allocate(rigidity(nnzmax))
    allocate(rowptr(list_length+1))
    allocate(colind(nnzmax))

    rigidity = 0._dp
    rowptr = 0
    colind = 0

    ! rigidity matrix in compressed row format
    row = 1
    nz = 1
    do i = 1,nat
        do j=marker1(i),marker2(i)
            bdry = .false.
            rowptr(row) = nz
            rij=at(i)%x-at(list(j))%x

            where (abs(rij)/samplesize > 0.5_dp .and. pbc)
                rij = rij - sign(samplesize,rij)
                bdry = .true.
            end where

            dist_ij = sqrt(dot_product(rij,rij))
            do k=1,2*dimn
                if (k .le. dimn) then
                    colind(nz) = dimn*(i-1)+k
                    rigidity(nz) = rij(k)/dist_ij
                else
                    colind(nz) = dimn*(list(j)-1)+(k-dimn)
                    rigidity(nz) = -rij(k-dimn)/dist_ij
                endif
                nz=nz+1
            enddo

            do k=1,dimn
                if (bdry(k) .eqv. .true.) then
                    colind(nz) = dimn*nat + k
                    rigidity(nz) = abs(rij(k))/dist_ij
                    nz = nz + 1
                endif
            enddo

            row = row+1
        enddo
        rowptr(row) = nz
    enddo

    end subroutine construct_rigidity

    subroutine project_forces
    !-------------------------------------------------------------------------------
    ! SUBROUTINE project_forces :
    !
    ! Passed Parameters :-
    ! -----------------
    !
    ! Non-Obvious Parameter :-
    ! ---------------------
    !
    use mod_output, only : string,output,erroroutput
    use mod_fij, only : fij
    use mod_neigh, only : list,marker1,marker2
    use lsqrModule,     only : LSQR
    use lsqrCheckModule,only : Acheck
    use mod_file, only : GetFreeUnit
    use mod_particles, only : at
    use mod_box, only : nat,dimn,samplesize,pbc
    implicit none

    integer  :: i,j,k,lsqrunit,istop,itn,inform,lsqitnlim
    real(dp) :: damp,eps,atol,btol,Acond,conlim,Anorm,rnorm,Arnorm,xnorm
    logical :: wantse
    real(kind=dp), dimension(:),allocatable :: x,se
    real(kind=dp) :: rij(dimn)

    write(string,'(a)')'Starting projected interatomic force calculation'
    call output
    call construct_rigidity

    allocate(proj_fij_scalar(nrow))
    allocate(proj_fij(dimn,nrow))
    allocate(se(ncol))
    allocate(x(ncol))
    allocate(fij_scalar(nrow))

    proj_fij = 0._dp
    proj_fij_scalar = 0._dp
    fij_scalar = 0._dp

    do i = 1,nat
        do k=marker1(i),marker2(i)
            j = list(k)
            rij = at(i)%x - at(j)%x
            where (abs(rij)/samplesize > 0.5_dp .and. pbc)
                rij = rij - sign(samplesize,rij)
            end where
            fij_scalar(k) = -dot_product(fij(:,k),rij/sqrt(dot_product(rij,rij)))
        enddo
    enddo

    write(string,'(a)')' '; call output
    write(string,'(a)')'Recasting the projection problem as a ' // &
                       'Linear Least Squares problem (LSQR) : '; call output
    write(string,'(a)')'min ||Ax-b||, where Cayley-Menger vectors ' // &
                       'are the columns of A, and b = fij_scalar'; call output
    write(string,'(a)')' '; call output
    write(string,'(a)')'Starting the LSQR solver'; call output
    write(string,'(a)')'========================================================'
    call output

    eps    = epsilon(eps)
    damp = 0._dp
    wantse = .true.
    atol   = eps**0.99_dp                           ! Asks for high accuracy.
    btol = atol
    conlim = 10000000._dp
    lsqitnlim = 4*ncol
    !-------------------------------------------------------------------------------
    ! open the LSQR.txt file
    !
    lsqrunit=6

    call Acheck(nrow,ncol,Aprod1,Aprod2,lsqrunit,inform)   ! Check Aprod1, Aprod2.
    if (inform > 0) then
      write(lsqrunit,'(a)') 'Check tol in subroutine Acheck in lsqrCheckModule'
      stop
    end if

    call LSQR  (nrow,ncol,Aprod1,Aprod2,fij_scalar,damp,wantse,         &
                x,se,atol,btol,conlim,lsqitnlim,lsqrunit,        &
                istop,itn,Anorm,Acond,rnorm,Arnorm,xnorm)
    write(string,'(a)')'========================================================'
    call output
    call Aprod1(nrow,ncol,x,proj_fij_scalar)
    call check
    do i=1,nat
        do k = marker1(i),marker2(i)
            j = list(k)
            rij = at(i)%x - at(j)%x
            where (abs(rij)/samplesize > 0.5_dp .and. pbc)
                rij = rij - sign(samplesize,rij)
            end where
            proj_fij(:,k) = -proj_fij_scalar(k) * rij/sqrt(dot_product(rij,rij))
        enddo
    enddo
            
    deallocate(x)
    deallocate(se)

    end subroutine project_forces

    subroutine Aprod1(m,n,x,y)
    use mod_qclib, only : amux
    implicit none
    ! m = nrow, n = ncol
    integer,  intent(in)    :: m,n
    real(kind=dp), intent(in)    :: x(n)
    real(kind=dp), intent(inout) :: y(m)

    real(kind=dp) :: z(m)
    ! Calculate y = Ax
    call amux(m,x,z,rigidity,colind,rowptr)
    y = y + z
    end subroutine Aprod1

    subroutine Aprod2(m,n,x,y)
    use mod_qclib, only : atmux
    implicit none
    integer,  intent(in)    :: m,n
    real(kind=dp), intent(inout)  :: x(n)
    real(kind=dp), intent(in) :: y(m)

    real(kind=dp) :: z(n)
    ! Calculate x = A'y
    call atmux(m,n,y,z,rigidity,colind,rowptr)
    x = x + z
    end subroutine Aprod2

    subroutine deallocate_rigidity
    implicit none
    if (allocated(rigidity)) deallocate(rigidity)
    if (allocated(colind)) deallocate(colind)
    if (allocated(rowptr)) deallocate(rowptr)
    if (allocated(proj_fij_scalar)) deallocate(proj_fij_scalar)
    if (allocated(fij_scalar)) deallocate(fij_scalar)
    if (allocated(proj_fij)) deallocate(proj_fij)
    end subroutine deallocate_rigidity

    subroutine check
    use mod_box, only         : dimn,nat,samplesize,pbc
    use mod_neigh, only       : marker1,marker2,list
    use mod_particles,   only : at           
    implicit none
    integer i,j,k
    real(kind=dp),dimension(dimn) :: rij
    real(kind=dp),dimension(dimn,nat) :: proj_force
    real(kind=dp) :: rsqij

    proj_force = 0.d0
    do i=1,nat
        do k = marker1(i),marker2(i)
            j = list(k)
            rij = at(i)%x - at(j)%x
            where (abs(rij)/samplesize > 0.5_dp .and. pbc)
                rij = rij - sign(samplesize,rij)
            end where
            rsqij = dot_product(rij,rij)
            proj_force(:,i) = proj_force(:,i) + (fij_scalar(k)-proj_fij_scalar(k))*rij/sqrt(rsqij)
            proj_force(:,j) = proj_force(:,j) - (fij_scalar(k)-proj_fij_scalar(k))*rij/sqrt(rsqij) 
            !proj_force(:,i) = proj_force(:,i) - proj_fij_scalar(k)*rij/sqrt(rsqij)
            !proj_force(:,j) = proj_force(:,j) + proj_fij_scalar(k)*rij/sqrt(rsqij) 
        enddo
    enddo
    !write(*,*) "Projected forces:"
    !do i=1,nat
    !    write(*,*) proj_force(:,i)
    !enddo
    !stop
    end subroutine check

end module mod_rigidity

